<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 1/29/16
 * Time: 8:03 AM
 */

namespace qfq;

require_once(__DIR__ . '/../store/Store.php');
require_once(__DIR__ . '/UserFormException.php');
require_once(__DIR__ . '/../Constants.php');
require_once(__DIR__ . '/../store/T3Info.php');


class AbstractException extends \Exception {

    public $messageArray = array();
    public $store;

    protected $file = '';
    protected $line = '';

    /**
     * @return string
     */
    public function formatException() {

        $debug = '';
        try {
            // In a very early stage, it might be possible that Store can't be initialized: take care not to use it.
            $store = Store::getInstance();
        } catch (UserFormException $e) {
            $store = null;
        } catch (CodeException $e) {
            $store = null;
        }

        $html = '';

        $this->messageArray['File'] = $this->getFile();
        $this->messageArray['Line'] = $this->getLine();
        $this->messageArray['Message'] = $this->getMessage();
        $this->messageArray['Code'] = $this->getCode();
        $this->messageArray['Timestamp'] = date('Y.m.d H:i:s O');

        //TODO: 'Client' will remove any HTML Tags before displaying. Formatting needs to be solved together with 'Client',
        // Respect mode: 'form load' (plain HTML), 'Ajax alert'
        $this->messageArray['Stacktrace'] = $this->getTraceAsString();
//        $this->messageArray['Stacktrace'] = $this->getTrace();

        $html .= "Code: " . htmlspecialchars($this->messageArray['Code']) . "<br>";
        $html .= "Message: " . Support::wrapTag("<strong>", htmlspecialchars($this->messageArray['Message'])) . "<br>";

        $t3Vars = T3Info::getVars();
        $html .= 'Page Id: ' . $t3Vars[TYPO3_PAGE_ID] . '<br>';
        $html .= 'Content Id: ' . $t3Vars[TYPO3_TT_CONTENT_UID] . '<br>';

        // Form
        if (isset($this->messageArray['Form']) && $this->messageArray['Form'] != '') {
            $html .= "Form: " . Support::wrapTag("<strong>", htmlspecialchars($this->messageArray['Form'])) . "<br>";
        }

        if (isset($this->messageArray['Form Element']) && $this->messageArray['Form Element'] != '') {
            $html .= "Form Element: " . Support::wrapTag("<strong>", htmlspecialchars($this->messageArray['Form Element'])) . "<br>";
        }

        // Report
        if (isset($this->messageArray[SYSTEM_REPORT_COLUMN_INDEX])) {
            $html .= "Column index: " . Support::wrapTag("<strong>", htmlspecialchars($this->messageArray[SYSTEM_REPORT_COLUMN_INDEX])) . "<br>";
        }

        if (isset($this->messageArray[SYSTEM_REPORT_COLUMN_NAME])) {
            $html .= "Column name: " . Support::wrapTag("<strong>", htmlspecialchars($this->messageArray[SYSTEM_REPORT_COLUMN_NAME])) . "<br>";
        }

        if (isset($this->messageArray[SYSTEM_REPORT_COLUMN_VALUE])) {
            $html .= "Column value: " . Support::wrapTag("<strong>", htmlspecialchars($this->messageArray[SYSTEM_REPORT_COLUMN_VALUE])) . "<br>";
        }

        $html = "<h2>Error</h2>" . Support::wrapTag('<p>', $html);
        $html = Support::wrapTag("<div class='warning'>", $html);

        if ($store !== null && $store->getVar(SYSTEM_SHOW_DEBUG_INFO, STORE_SYSTEM) === 'yes') {

            $this->messageArray['current sip'] = $store->getStore(STORE_SIP);

            // Layout
            $debug = '<tr bgcolor="#dddddd"><td colspan="2">Exception</td></tr>';
            foreach ($this->messageArray as $key => $value) {
                if (is_array($value)) {
                    $value = var_export($value, true);
                }

                if ($value !== '' && $value !== false)
                    $debug .= "<tr>" . "<td>$key</td>" . "<td>" . htmlspecialchars($value) . "</td>" . "</tr>";
            }
            $debug = "<table border=1>" . $debug . "</table>";
        }
        return $html . $debug;
    }

}