<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 4/17/17
 * Time: 10:17 PM
 *
 * Check: CODING.md > 'Print' and 'Download'
 *
 */

namespace qfq;

require_once(__DIR__ . '/../store/Config.php');
require_once(__DIR__ . '/../Constants.php');
require_once(__DIR__ . '/../helper/KeyValueStringParser.php');


class Html2Pdf {

    /**
     * @var array
     */
    private $config = array();

    /**
     * Read QFQ config. Only SYSTEM_BASE_URL_PRINT and SYSTEM_WKHTMLTOPDF will be used.
     * Check and get all clean _GET Parameter. Build a URL based on SYSTEM_BASE_URL_PRINT and the delivered URL params.
     *
     * @param array $config
     * @throws UserFormException
     * @throws \exception
     */
    public function __construct(array $config = array()) {

        if (count($config) == 0) {
            $cfg = new Config();

            $config = $cfg->readConfig('');
        }

        $this->config = $config;

        if (!isset($config[SYSTEM_BASE_URL_PRINT]) || $config[SYSTEM_BASE_URL_PRINT] == '') {
            throw new \exception(CONFIG_INI . ' - Missing ' . SYSTEM_BASE_URL_PRINT);
        }

        if (!isset($config[SYSTEM_WKHTMLTOPDF]) || $config[SYSTEM_WKHTMLTOPDF] == '') {
            throw new \exception(CONFIG_INI . ' - Missing ' . SYSTEM_WKHTMLTOPDF);
        }

        if (!is_executable($config[SYSTEM_WKHTMLTOPDF])) {
            throw new \exception(CONFIG_INI . ' - ' . SYSTEM_WKHTMLTOPDF . '=' . $config[SYSTEM_WKHTMLTOPDF] . ' - not found or not executable.');
        }
    }

    /**
     * Return an array with GET params who are clean - they do not violate $pattern.
     *
     * @return array
     */
    private function readCleanGetParam(array $get) {

        $param = array();
        $pattern = '^[\-_\.,;:\/a-zA-Z0-9]*$'; // ':alnum:' does not work here in FF

        foreach ($get as $key => $value) {
            if (preg_match("/$pattern/", $value) === 1) {
                $param[$key] = $value;
            }
        }

        return $param;
    }

    /**
     * Set HTML Header to initiate PDF download.
     *
     * @param $filename
     */
    private function setHeader($filename) {

        header("Content-Disposition: inline; filename=\"$filename\"");
        header("Content-Type: application/pdf");
        header("Content-Transfer-Encoding: binary");
    }

    /**
     * Converts a Webpage (URL) to a PDF file.
     * The URL might be a local Typo3 page (without hostname, starting with the parameter) or a full URL.
     *
     * @param string $token TOKEN_URL | TOKEN_URL_PARAM
     * @param string $url id=exportPage&r=123, www.nzz.ch/issue?id=456
     * @return string        rendered file - please delete later
     * @throws \exception
     */
    public function page2pdf($token, $url) {
        $args = array();
        $urlParamNew = array();
        $urlParamString = '';
        $host = '';

        switch ($token) {
            case TOKEN_URL:
                $arr = explode('?', $url, 2);
                $host = $arr[0];
                $urlParamString = empty($arr[1]) ? '' : $arr[1];
                break;
            case TOKEN_URL_PARAM:
                $host = $this->config[SYSTEM_BASE_URL_PRINT];
                $urlParamString = $url;
//                $url = $this->config[SYSTEM_BASE_URL_PRINT] . '/?' . $url;
                break;
            default:
                break;
        }

        $urlParam = KeyValueStringParser::parse($urlParamString, '=', '&', KVP_IF_VALUE_EMPTY_COPY_KEY);
        foreach ($urlParam as $key => $value) {
            if (substr($key, 0, 1) == '-') {
                $args[$key] = $value;
            } else {
                $urlParamNew[$key] = $value;
            }
        }

        $args = OnArray::arrayEscapeshellarg($args);
        $options = KeyValueStringParser::unparse($args, ' ', ' ');
        $urlParamString = KeyValueStringParser::unparse($urlParamNew, '=', '&');
        if ($urlParamString != '') {
            if (substr($host, -1, 1) != '/') {
                $host .= '/';
            }
            $url = $host . '?' . $urlParamString;
        }

        if (substr($url, 0, 4) != 'http') {
            $url = 'http://' . $url;
        }

        $urlPrint = escapeshellarg($url);
        $wkhtmlToPdf = $this->config[SYSTEM_WKHTMLTOPDF];

        $filename = tempnam(sys_get_temp_dir(), DOWNLOAD_FILE_PREFIX);
        $filenameEscape = escapeshellarg($filename);

        $cmd = "$wkhtmlToPdf $options $urlPrint $filenameEscape";

        $rc = 0;
        $line = system($cmd, $rc);

        if ($rc != 0) {
            throw new \exception("Error [RC=$rc] $line: $cmd");
        }

        return $filename;
    }

    /**
     * @throws \exception
     */
    public function outputHtml2Pdf() {

        $get = $this->readCleanGetParam($_GET);
        $urlParam = KeyValueStringParser::unparse($get, '=', '&');
        $pageId = Support::setIfNotSet($get, HTML2PDF_PAGEID, 0);

        $filename = $this->page2pdf(TOKEN_URL_PARAM, $urlParam);

        $this->setHeader('print.' . $pageId . '.pdf');
        @readfile($filename);
        @unlink($filename);
//        @unlink($filename . '.log');

        exit; // Do an extremely hard exit here to make sure there are no more additional bytes sent (makes the delivered PDF unusable).

    }

}